@@ -1,7 +1,17 @@
 source("incl/start.R")
-
 options(future.debug = FALSE)
 
+f_try <- function(f, ...) {
+  res <- tryCatch(..., error = identity)
+  if (inherits(res, "error")) {
+    label <- sQuote(f$label)
+    if (length(label) == 0) label <- "<none>"
+    msg <- sprintf("Test with future (%s) of class %s failed assertion: %s", label, sQuote(class(f)[1]), conditionMessage(res))
+    stop(msg, call.=FALSE)
+  }
+  res
+}
+
 message("*** immediateCondition:s ...")
 
 immediateMessage <- function(..., domain = NULL, appendLF = TRUE) {
@@ -35,7 +45,17 @@ recordMessages <- function(expr, ...) {
   sapply(recordConditions(expr, ...), FUN = conditionMessage)
 }
 
-strategies <- supportedStrategies()
+## FIXME: Make sure to set also with cores = 2L /HB 2020-11-10
+
+if (isTRUE(getOption("future.psock.relay.immediate"))) {
+  excl <- "cluster"
+  if (getRversion() < "3.4.0") excl <- c(excl, "multisession")
+  strategies <- supportedStrategies(cores = 2L, excl = excl)
+  strategies <- c(strategies, "sequential")
+} else {
+  strategies <- "sequential"
+}
+print(strategies)
 
 for (ss in seq_along(strategies)) {
   strategy <- strategies[[ss]]
@@ -51,16 +71,22 @@ for (ss in seq_along(strategies)) {
       message("M")
       immediateWarning("IW")
       immediateMessage("IM2")
+      ## RACE CONDITION IN ASSERTION:
+      ## Add some leeway for the immediateCondition files written by
+      ## multicore futures to "settle" on the file system
+      Sys.sleep(1.0)
       42L
-    })
+    }, label = "single-future")
   })
   message("  class: ", paste(sQuote(class(f)), collapse = ", "))
   message(sprintf("  msgs [n=%d]: %s", length(msgs), paste(sQuote(msgs), collapse = ", ")))
-  if (inherits(f, "UniprocessFuture")) {
-    stopifnot(identical(msgs, c("IM1\n", "IW", "IM2\n")))
-  } else {
-    stopifnot(length(msgs) == 0L)
-  }
+  f_try(f, {
+    if (inherits(f, "UniprocessFuture")) {
+      stopifnot(identical(msgs, c("IM1\n", "IW", "IM2\n")))
+    } else {
+      stopifnot(length(msgs) == 0L)
+    }
+  })
   
   message("- checking if resolved")
   msgs <- recordMessages({
@@ -68,70 +94,99 @@ for (ss in seq_along(strategies)) {
   })
   message("  result: ", r)
   message(sprintf("  msgs [n=%d]: %s", length(msgs), paste(sQuote(msgs), collapse = ", ")))
-  if (inherits(f, "ClusterFuture")) {
-    stopifnot(identical(msgs, c("IM1\n")))
-  } else {
-    stopifnot(length(msgs) == 0L)
-  }
+  f_try(f, {
+    if (inherits(f, "MultiprocessFuture")) {
+      stopifnot(identical(msgs, c("IM1\n", "IW", "IM2\n")))
+    } else {
+      stopifnot(length(msgs) == 0L)
+    }
+  })
 
   message("- resolve w/out collecting results")
   msgs <- recordMessages({
     f <- resolve(f)
   })
   message(sprintf("  msgs [n=%d]: %s", length(msgs), paste(sQuote(msgs), collapse = ", ")))
-  if (inherits(f, "ClusterFuture")) {
-    stopifnot(identical(msgs, c("IW", "IM2\n")))
-  } else {
+  f_try(f, {
     stopifnot(length(msgs) == 0L)
-  }
-
+  })
+  
   message("- resolve w/ collect results")
   msgs <- recordMessages({
     f <- resolve(f, result = TRUE)
   })
   message(sprintf("  msgs [n=%d]: %s", length(msgs), paste(sQuote(msgs), collapse = ", ")))
-  if (inherits(f, c("UniprocessFuture", "ClusterFuture", "CallrFuture", "BatchtoolsFuture"))) {
-    stopifnot(length(msgs) == 0L)
-  } else {
-    stopifnot(identical(msgs, c("IM1\n", "IW", "IM2\n")))
-  }
-
+  f_try(f, {
+    if (inherits(f, c("UniprocessFuture", "MultiprocessFuture"))) {
+      stopifnot(length(msgs) == 0L)
+    } else {
+      stopifnot(identical(msgs, c("IM1\n", "IW", "IM2\n")))
+    }
+  })
+  
   message("- getting value")
   msgs <- recordMessages({
     v <- value(f)
   })
   message("  value: ", v)
   message(sprintf("  msgs [n=%d]: %s", length(msgs), paste(sQuote(msgs), collapse = ", ")))
-  if (inherits(f, c("UniprocessFuture", "ClusterFuture"))) {
-    stopifnot(identical(msgs, "M\n"))
-  } else {
-    stopifnot(identical(msgs, c("IM1\n", "IW", "IM2\n", "M\n")))
-  }
+  f_try(f, {
+    if (inherits(f, c("UniprocessFuture", "MultiprocessFuture"))) {
+      stopifnot(identical(msgs, "M\n"))
+    } else {
+      stopifnot(identical(msgs, c("IM1\n", "IW", "IM2\n", "M\n")))
+    }
+  })
   
   message("- getting value again")
   msgs <- recordMessages({
     v <- value(f)
   })
   message("  value: ", v)
-  stopifnot(identical(msgs, "M\n"))
+  f_try(f, {
+    stopifnot(identical(msgs, "M\n"))
+  })
 
   message("* A single future ... DONE")
 
+
   message("* Two futures ... ")
 
   message("- list of two futures")
   fs <- list()
   msgs <- recordMessages({
-    fs[[1]] <- future({ immediateMessage("IM1"); Sys.sleep(0.1); message("M1"); immediateWarning("IW1"); 1L })
-    fs[[2]] <- future({ immediateMessage("IM2"); Sys.sleep(0.1); message("M2"); immediateWarning("IW2"); 2L })
+    fs[[1]] <- future({
+      immediateMessage("IM1")
+      Sys.sleep(0.1)
+      message("M1")
+      immediateWarning("IW1")
+      ## RACE CONDITION IN ASSERTION:
+      ## Add some leeway for the immediateCondition files written by
+      ## multicore futures to "settle" on the file system
+      Sys.sleep(1.0)
+      1L
+    }, label = "future-1")
+    fs[[2]] <- future({
+      immediateMessage("IM2")
+      Sys.sleep(0.1)
+      message("M2")
+      immediateWarning("IW2")
+      ## RACE CONDITION IN ASSERTION:
+      ## Add some leeway for the immediateCondition files written by
+      ## multicore futures to "settle" on the file system
+      Sys.sleep(1.0)
+      2L
+    }, label = "future-2")
   })
   message("  class: ", paste(sQuote(class(fs[[1]])), collapse = ", "))
   message(sprintf("  msgs [n=%d]: %s", length(msgs), paste(sQuote(msgs), collapse = ", ")))
-  if (inherits(fs[[1]], "UniprocessFuture")) {
-    stopifnot(identical(msgs, c("IM1\n", "IW1", "IM2\n", "IW2")))
-  } else {
-    stopifnot(length(msgs) == 0L)
-  }
+  f_try(fs[[1]], {
+    if (inherits(fs[[1]], "UniprocessFuture")) {
+      stopifnot(identical(msgs, c("IM1\n", "IW1", "IM2\n", "IW2")))
+    } else {
+      stopifnot(length(msgs) == 0L)
+    }
+  })
 
   message("- check if resolved")
   msgs <- recordMessages({
@@ -139,33 +194,40 @@ for (ss in seq_along(strategies)) {
   })
   message("  result: ", paste(rs, collapse = ", "))
   message(sprintf("  msgs [n=%d]: %s", length(msgs), paste(sQuote(msgs), collapse = ", ")))
-  if (inherits(f, "ClusterFuture")) {
-    stopifnot(identical(msgs, c("IM1\n", "IM2\n")))
-  } else {
-    stopifnot(length(msgs) == 0L)
-  }  
+  f_try(f, {
+    if (inherits(f, "MultiprocessFuture")) {
+      stopifnot(all(msgs %in% c("IM1\n", "IM2\n", "IW1", "IW2")))
+      if (inherits(f, "MulticoreFuture")) {
+        stopifnot(identical(msgs, c("IM1\n", "IM2\n", "IW1", "IW2")))
+      } else if (inherits(f, "MultisessionFuture")) {
+        stopifnot(identical(msgs, c("IM1\n", "IW1", "IM2\n", "IW2")))
+      }
+    } else {
+      stopifnot(length(msgs) == 0L)
+    }
+  })
 
   message("- resolve w/out collecting results")
   msgs <- recordMessages({
     fs <- resolve(fs)
   })
   message(sprintf("  msgs [n=%d]: %s", length(msgs), paste(sQuote(msgs), collapse = ", ")))
-  if (inherits(f, "ClusterFuture")) {
-    stopifnot(identical(msgs, c("IW1", "IW2")))
-  } else {
+  f_try(f, {
     stopifnot(length(msgs) == 0L)
-  }  
+  })
 
   message("- resolve w/ collect results")
   msgs <- recordMessages({
     fs <- resolve(fs, result = TRUE)
   })
   message(sprintf("  msgs [n=%d]: %s", length(msgs), paste(sQuote(msgs), collapse = ", ")))
-  if (inherits(fs[[1]], c("UniprocessFuture", "ClusterFuture", "CallrFuture", "BatchtoolsFuture"))) {
-    stopifnot(length(msgs) == 0L)
-  } else {
-    stopifnot(identical(msgs, c("IM1\n", "IW1", "IM2\n", "IW2")))
-  }
+  f_try(f, {
+    if (inherits(fs[[1]], c("UniprocessFuture", "MultiprocessFuture", "CallrFuture", "BatchtoolsFuture"))) {
+      stopifnot(length(msgs) == 0L)
+    } else {
+      stopifnot(identical(msgs, c("IM1\n", "IW1", "IM2\n", "IW2")))
+    }
+  })
 
   message("- getting value")
   msgs <- recordMessages({
@@ -173,7 +235,9 @@ for (ss in seq_along(strategies)) {
   })
   message("  values: ", paste(vs, collapse = ", "))
   message(sprintf("  msgs [n=%d]: %s", length(msgs), paste(sQuote(msgs), collapse = ", ")))
-  stopifnot(identical(msgs, c("M1\n", "M2\n")))
+  f_try(f, {
+    stopifnot(identical(msgs, c("M1\n", "M2\n")))
+  })
   
   message("- getting value again")
   msgs <- recordMessages({
@@ -181,7 +245,9 @@ for (ss in seq_along(strategies)) {
   })
   message("  values: ", paste(vs, collapse = ", "))
   message(sprintf("  msgs [n=%d]: %s", length(msgs), paste(sQuote(msgs), collapse = ", ")))
-  stopifnot(identical(msgs, c("M1\n", "M2\n")))
+  f_try(f, {
+    stopifnot(identical(msgs, c("M1\n", "M2\n")))
+  })
 
   message("- getting values one by one")
   msgs <- recordMessages({
@@ -191,7 +257,9 @@ for (ss in seq_along(strategies)) {
     )
   })
   message("  values: ", paste(vs, collapse = ", "))
-  stopifnot(identical(msgs, c("M1\n", "M2\n")))
+  f_try(f, {
+    stopifnot(identical(msgs, c("M1\n", "M2\n")))
+  })
 
   message("* Two futures ... DONE")
 
